/* pthread_mutex_main.c
 *
 * pthreads / with mutex synchronisation compile with -DMUTEX
 * main releases mutex after thread creation
 * simple error handling
 *
 * Bernhard Rabe
 * 17.11.2004
 */


#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

/*global accessible*/
pthread_mutex_t mutex; 

/* Thread routine */
void* run(void* data)
{
    int id=*(int*)data,l;
   
#ifdef MUTEX    
    if(0!=pthread_mutex_lock(&mutex)) { fprintf(stderr,"mutex_lock\n");exit(1); }
#endif    
    for(l=0;l<10;l++)
    {
        printf("%c",id);
        usleep(10);
    }
    printf("\n");
#ifdef MUTEX    
    if(0!=pthread_mutex_unlock(&mutex)) { fprintf(stderr,"mutex_unlock\n");exit(1);}
#endif    
    pthread_exit((void*)0);
}


int main(int argc, char *argv[])
{
    pthread_t thread[10];
    int ids[10]={'-','*','+','#','~','$','','%','^',''},i=2,l;
    
    if(argc>1)
    {
        i=atoi(argv[1]);
        if(i<2 || i>10)
        {
            printf("usage: %s [2-10]\n",argv[0]);
            exit(1);
        }
    }
#ifdef MUTEX    
    /*create mutex, for access to stdout*/
    if(0!=pthread_mutex_init(&mutex,NULL))
    {/*Fehler*/ 
        fprintf(stderr,"mutex_init\n");exit(1);
    }
    /* get the mutex */
    if(0!=pthread_mutex_lock(&mutex))
    {/*Fehler*/ 
        fprintf(stderr,"mutex_lock\n");exit(1);
    }
#endif    
    /*create and start threads*/  
    for(l=0;l<i;l++)
        if(0!=pthread_create(&thread[l],0,run,&ids[l]))
        {
            fprintf(stderr,"pthread_create\n");exit(1);
        }
    
        
    printf("Waiting for Threads\n"); 
#ifdef MUTEX   
    /*release Mutex*/        
    if(0!=pthread_mutex_unlock(&mutex))
    {/*Fehler*/ 
        fprintf(stderr,"mutex_init\n");exit(1);
    }
#endif    
    
    /* wait for thread termination, must be in joinable mode */ 
    for(l=0;l<i;l++)
        if(0!=pthread_join(thread[l],0))
        { 
            fprintf(stderr,"pthread_join\n");exit(1);
        }
    
#ifdef MUTEX   
    if(0!=pthread_mutex_destroy(&mutex))
    { 
            fprintf(stderr,"mutex_destroy\n");exit(1);
    }    
#endif    
    return 0;
}
