/*
 * pthread_prime.c
 * POSIX Threads to solve practice 2.4
 * only minimal error handling
 *
 * Bernhard Rabe
 * 22.11.2004
 */

#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <limits.h>

int is_prime(int n)
{
    int i;

    if (n < 2) return 0;

    for (i=2; i*i <= n; i++)
    {
        if ((n % i) == 0)
            return 0;
    }

    return 1;
}

#include <math.h>

double f (double n)
{
        return exp (log (n) * (3.0 / 2.0));
}

double f_reverse (double n)
{
        return exp (log (n) * (2.0 / 3.0));
}

typedef struct {
    int from;
    int to;
} bounds;    

void *run(void *data)
{
    int i, *teilsumme;
    bounds *bd=(bounds*)data;
    
    if(NULL==(teilsumme=(int*)malloc(sizeof(int))))
    {
        perror("malloc");
        exit(1);
    }
    
    *teilsumme=0;

    for (i=bd->from; i <= bd->to; i++)
    {	
	if (is_prime(i))
	{
            *teilsumme +=  i%100;
            *teilsumme %= 100;
        }
    }
    pthread_exit((void*)(teilsumme));
}


int main(int argc, char **argv)
{
  int anzahl_threads;
  int max_primzahl;
  int sum=0,i;
  int von;
  int bis;
  double cycles;
  pthread_t *pth;
  bounds *bd;
    
  if(argc<3)
  {
      printf("usage: %s maxprime maxthreads\n",argv[0]);
      exit(1);
  }

  max_primzahl=atoi(argv[1]);
  anzahl_threads=atoi(argv[2]);
  
  if(max_primzahl<2 || max_primzahl>INT_MAX || anzahl_threads<1)
  {
      printf("maxprime=[2-%d] maxthreads>0\n",INT_MAX);
      exit(1);      
  }
  von = 0;
  bis = 0;
  cycles = f(max_primzahl);
  if(NULL==(pth=(pthread_t*)malloc(sizeof(pthread_t)*anzahl_threads)))
  {
      perror("malloc");
      exit(1);
  }
    
  if(NULL==(bd=(bounds*)malloc(sizeof(bounds)*anzahl_threads)))
  {
      perror("malloc");
      exit(1);
  }
  for (i=0; i < anzahl_threads; i++)
  {
     von = bis + 1;
     bis = f_reverse((i + 1) * cycles / anzahl_threads);
     
     if (i == (anzahl_threads - 1)) /* letzter Prozess */
     {
	 bis = max_primzahl;
     }
     if (bis < von)
     {
	printf("zu viele Prozesse fr zu wenig Zahlen: %d %d %d\n",i,von,bis);
	exit(1);
     }
     bd[i].from=von;
     bd[i].to=bis;     
     if(0!=pthread_create(&pth[i],0,run,&bd[i]))
     {
         fprintf(stderr,"pthread_create\n");exit(1);
     }
  }
  
  for (i=0; i < anzahl_threads; i++)
  {
      int *pret;
      
      if(0!=pthread_join(pth[i],(void**)&pret))
      {
          fprintf(stderr,"pthread_join\n");exit(1);
      }
      
      sum+= *pret;
      sum%=100; 

      free(pret);
  }    
  printf("%d\n",sum);
  
  free(bd);/*Thread boundaries*/
  free(pth);/*Thread Id's*/
  
  return 0;
}
